@@ -34,6 +34,7 @@
 #endif /* GLOBAL_CONFIG_FILE */
 
 extern char **environ;
+char **g_argv;
 
 #define ISACTIVE (appstate.active)
 #define ISDRAGGING (appstate.dragging)
@@ -96,6 +97,7 @@ static viewport_t *viewports;
 static int nviewports = 0;
 static int xinerama = 0;
 static int daemonize = 0;
+static int is_daemon = False;
 
 static Display *dpy;
 static Window zone;
@@ -127,36 +129,40 @@ static wininfo_t wininfo_history[WININFO_MAXHIST]; /* XXX: is 100 enough? */
 static int wininfo_history_cursor = 0;
 
 void defaults();
+
 void cmd_cell_select(char *args);
 void cmd_click(char *args);
+void cmd_cursorzoom(char *args);
 void cmd_cut_down(char *args);
 void cmd_cut_left(char *args);
 void cmd_cut_right(char *args);
 void cmd_cut_up(char *args);
+void cmd_daemonize(char *args);
 void cmd_doubleclick(char *args);
 void cmd_drag(char *args);
 void cmd_end(char *args);
 void cmd_grid(char *args);
 void cmd_grid_nav(char *args);
 void cmd_history_back(char *args);
+void cmd_loadconfig(char *args);
 void cmd_move_down(char *args);
 void cmd_move_left(char *args);
 void cmd_move_right(char *args);
 void cmd_move_up(char *args);
-void cmd_cursorzoom(char *args);
-void cmd_windowzoom(char *args);
 void cmd_quit(char *args);
+void cmd_record(char *args);
+void cmd_restart(char *args);
 void cmd_shell(char *args);
 void cmd_start(char *args);
 void cmd_warp(char *args);
-void cmd_record(char *args);
+void cmd_windowzoom(char *args);
 
 void update();
 void correct_overflow();
 void handle_keypress(XKeyEvent *e);
 void handle_commands(char *commands);
 void parse_config();
-void parse_config_line(char *line);
+int parse_config_line(char *line);
 void save_history_point();
 void restore_history_point(int moves_ago);
 void cell_select(int x, int y);
@@ -173,6 +179,8 @@ void viewport_right();
 int pointinrect(int px, int py, int rx, int ry, int rw, int rh);
 int percent_of(int num, char *args, float default_val);
 void sigchld(int sig);
+void sighup(int sig);
+void restart();
 void recordings_save(const char *filename);
 void parse_recordings(const char *filename);
 void openpixel(Display *dpy, Window zone, mouseinfo_t *mouseinfo);
@@ -207,11 +215,14 @@ dispatch_t dispatch[] = {
   "drag", cmd_drag,
 
   // Other commands.
+  "loadconfig", cmd_loadconfig,
+  "daemonize", cmd_daemonize,
   "sh", cmd_shell,
   "start", cmd_start,
   "end", cmd_end, 
   "history-back", cmd_history_back,
   "quit", cmd_quit,
+  "restart", cmd_restart,
   "record", cmd_record,
   NULL, NULL,
 };
@@ -230,8 +241,8 @@ int parse_keycode(char *keyseq) {
   char *tok;
   char *last_tok;
   char *dup;
-  int keycode;
-  int keysym;
+  int keycode = 0;
+  int keysym = 0;
 
   strptr = dup = strdup(keyseq);
   //printf("finding keycode for %s\n", keyseq);
@@ -241,11 +252,17 @@ int parse_keycode(char *keyseq) {
   }
 
   keysym = XStringToKeysym(last_tok);
-  if (keysym == NoSymbol)
-    fprintf(stderr, "No keysym found for %s\n", last_tok);
-  keycode = XKeysymToKeycode(dpy, keysym);
-  if (keycode == 0)
-    fprintf(stderr, "Unable to lookup keycode for %s\n", last_tok);
+  if (keysym == NoSymbol) {
+    fprintf(stderr, "No keysym found for '%s' in sequence '%s'\n",
+            last_tok, keyseq);
+    /* At this point, we'll be returning 0 for keycode */
+  } else {
+    /* Valid keysym */
+    keycode = XKeysymToKeycode(dpy, keysym);
+    if (keycode == 0) {
+      fprintf(stderr, "Unable to lookup keycode for %s\n", last_tok);
+    }
+  }
 
   free(dup);
   return keycode;
@@ -370,14 +387,42 @@ void parse_config_file(const char* file) {
   FILE *fp = NULL;
 #define LINEBUF_SIZE 512
   char line[LINEBUF_SIZE];
+  int lineno = 0;
+
+  if (file[0] == '~') {
+    const char *homedir = getenv("HOME");
+
+    if (homedir != NULL) {
+      char *rcfile = NULL;
+      asprintf(&rcfile, "%s/%s", homedir, file + 1 /* skip first char '~' */);
+      parse_config_file(rcfile);
+      free(rcfile);
+      return;
+    } else {
+      fprintf(stderr, 
+              "No HOME set in environment. Can't expand '%s' (fatal error)\n",
+              file);
+      /* This is fatal. */
+      exit(1);
+    }
+  } /* if file[0] == '~' */
+
   fp = fopen(file, "r");
-  if (fp == NULL)
+
+  /* Silently ignore file read errors */
+  if (fp == NULL) {
+    //fprintf(stderr, "Error trying to open file for read '%s'\n", file);
+    //perror("Error");
     return;
+  }
   /* fopen succeeded */
   while (fgets(line, LINEBUF_SIZE, fp) != NULL) {
+    lineno++;
     /* Kill the newline */
     *(line + strlen(line) - 1) = '\0';
-    parse_config_line(line);
+    if (parse_config_line(line) != 0) {
+      fprintf(stderr, "Error with config %s:%d: %s\n", file, lineno, line);
+    }
   }
   fclose(fp);
 }
@@ -390,14 +435,7 @@ void parse_config() {
 
   defaults();
   parse_config_file(GLOBAL_CONFIG_FILE);
-  homedir = getenv("HOME");
-
-  if (homedir != NULL) {
-    char *rcfile = NULL;
-    asprintf(&rcfile, "%s/.keynavrc", homedir);
-    parse_config_file(rcfile);
-    free(rcfile);
-  }
+  parse_config_file("~/.keynavrc");
 }
 
 void defaults() {
@@ -419,6 +457,7 @@ void defaults() {
     "shift+k move-up",
     "shift+l move-right",
     "space warp,click 1,end",
+    "Return warp,click 1,end",
     "semicolon warp,end",
     "w warp",
     "t windowzoom",
@@ -447,12 +486,14 @@ void defaults() {
   };
   for (i = 0; default_config[i]; i++) {
     tmp = strdup(default_config[i]);
-    parse_config_line(tmp);
+    if (parse_config_line(tmp) != 0) {
+      fprintf(stderr, "Error with default config line %d: %s\n", i + 1, tmp);
+    }
     free(tmp);
   }
 }
 
-void parse_config_line(char *line) {
+int parse_config_line(char *orig_line) {
   /* syntax:
    * keysequence cmd1,cmd2,cmd3
    *
@@ -462,9 +503,9 @@ void parse_config_line(char *line) {
    * semicolon warp,click
    */
 
+  char *line = strdup(orig_line);
   char *tokctx;
   char *keyseq;
-  char *commands;
   int keycode, mods;
 
   char *comment;
@@ -480,28 +521,43 @@ void parse_config_line(char *line) {
 
   /* Ignore empty lines */
   if (*line == '\n' || *line == '\0')
-    return;
+    return 0;
 
   tokctx = line;
   keyseq = strdup(strtok_r(line, " ", &tokctx));
-  commands = strdup(tokctx);
 
   /* A special config option that will clear all keybindings */
   if (strcmp(keyseq, "clear") == 0) {
+    /* TODO(sissel): Make this a cmd_clear function */
     /* Reset keybindings */
     g_ptr_array_free(keybindings, TRUE);
     keybindings = g_ptr_array_new();
   } else if (strcmp(keyseq, "daemonize") == 0) {
-    daemonize = 1;
+    handle_commands(keyseq);
+  } else if (strcmp(keyseq, "loadconfig") == 0) {
+    handle_commands(keyseq);
   } else {
     keycode = parse_keycode(keyseq);
+    if (keycode == 0) {
+      fprintf(stderr, "Problem parsing keysequence '%s'\n", keyseq);
+      return 1;
+    }
     mods = parse_mods(keyseq);
 
-    addbinding(keycode, mods, commands);
+    /* FreeBSD sets 'tokctx' to NULL at end of string.
+     * glibc sets 'tokctx' to the next character (the '\0') 
+     * Reported by Richard Kolkovich */
+    if (tokctx == NULL || *tokctx == '\0') {
+      fprintf(stderr, "Incomplete configuration line. Missing commands: '%s'\n", line);
+      return 1;
+    }
+
+    addbinding(keycode, mods, tokctx /* the remainder of the line */);
   }
 
   free(keyseq);
-  free(commands);
+  free(line);
+  return 0;
 }
 
 int percent_of(int num, char *args, float default_val) {
@@ -513,6 +569,7 @@ int percent_of(int num, char *args, float default_val) {
   if (sscanf(args, "%f", &pct) <= 0)
     pct = default_val;
 
+  /* > 1, then it's not a percent, it's an absolute value. */
   if (pct > 1.0)
     return (int)pct;
 
@@ -855,6 +912,16 @@ void cmd_history_back(char *args) {
   restore_history_point(1);
 }
 
+void cmd_loadconfig(char *args) {
+  // Trim leading and trailing quotes if they exist
+  if (*args == '"') {
+    args++;
+    *(args + strlen(args) - 1) = '\0';
+  }
+
+  parse_config_file(args);
+}
+
 void cmd_shell(char *args) {
   // Trim leading and trailing quotes if they exist
   if (*args == '"') {
@@ -878,6 +945,10 @@ void cmd_quit(char *args) {
   exit(0);
 }
 
+void cmd_restart(char *args) {
+  restart();
+}
+
 void cmd_cut_up(char *args) {
   if (!ISACTIVE)
     return;
@@ -937,7 +1008,15 @@ void cmd_cursorzoom(char *args) {
   if (!ISACTIVE)
     return;
 
-  sscanf(args, "%d %d", &width, &height);
+  int count = sscanf(args, "%d %d", &width, &height);
+  if (count == 0) {
+    fprintf(stderr,
+            "Invalid usage of 'cursorzoom' (expected at least 1 argument)\n");
+  } else if (count == 1) {
+    /* If only one argument, assume we want a square. */
+    height = width;
+  }
+
   xdo_mouselocation(xdo, &xloc, &yloc, NULL);
 
   wininfo.x = xloc - (width / 2);
@@ -967,8 +1046,25 @@ void cmd_windowzoom(char *args) {
 void cmd_warp(char *args) {
   if (!ISACTIVE)
     return;
-  xdo_mousemove(xdo, wininfo.x + wininfo.w / 2, wininfo.y + wininfo.h / 2,
-                viewports[wininfo.curviewport].screen_num);
+  int x, y;
+  x = wininfo.x + wininfo.w / 2;
+  y = wininfo.y + wininfo.h / 2;
+  
+  if (mouseinfo.x != -1 && mouseinfo.y != -1) {
+    closepixel(dpy, zone, &mouseinfo);
+  }
+
+  /* Open pixels hould be relative to the window coordinates,
+   * not screen coordinates. */
+  mouseinfo.x = x - wininfo.x;
+  mouseinfo.y = y - wininfo.y;
+  openpixel(dpy, zone, &mouseinfo);
+
+  xdo_mousemove(xdo, x, y, viewports[wininfo.curviewport].screen_num);
+  xdo_mouse_wait_for_move_to(xdo, x, y);
+
+  /* TODO(sissel): do we need to open again? */
+  openpixel(dpy, zone, &mouseinfo);
 }
 
 void cmd_click(char *args) {
@@ -1018,6 +1114,7 @@ void cmd_drag(char *args) {
     appstate.dragging = False;
     xdo_mouseup(xdo, CURRENTWINDOW, button);
   } else { /* Start dragging */
+    cmd_warp(NULL);
     appstate.dragging = True;
     xdo_keysequence_down(xdo, 0, drag_modkeys, 12000);
     xdo_mousedown(xdo, CURRENTWINDOW, button);
@@ -1026,6 +1123,7 @@ void cmd_drag(char *args) {
     /* TODO(sissel): Make this a 'mousewiggle' command */
     xdo_mousemove_relative(xdo, 1, 0);
     xdo_mousemove_relative(xdo, -1, 0);
+    XSync(xdo->xdpy, 0);
     xdo_keysequence_up(xdo, 0, drag_modkeys, 12000);
   }
 }
@@ -1123,6 +1221,12 @@ void cell_select(int col, int row) {
   wininfo.y = wininfo.y + (wininfo.h * (row));
 }
 
+void cmd_daemonize(char *args) {
+  if (!is_daemon) {
+    daemonize = 1;
+  }
+}
+
 void cmd_record(char *args) {
   char *filename;
   if (!ISACTIVE)
@@ -1147,6 +1251,7 @@ void update() {
   if (!ISACTIVE)
     return;
 
+  /* Fix keynav window boundaries if they exceed the screen */
   correct_overflow();
   if (wininfo.w <= 1 || wininfo.h <= 1) {
     cmd_end(NULL);
@@ -1467,11 +1572,6 @@ void handle_commands(char *commands) {
         found = 1;
         dispatch[i].func(args);
 
-        if (ISDRAGGING)
-          cmd_warp(NULL);
-
-        /* Fix keynav window boundaries if they exceed the screen */
-        correct_overflow();
       }
     }
 
@@ -1480,6 +1580,8 @@ void handle_commands(char *commands) {
   }
 
   if (ISACTIVE) {
+    if (ISDRAGGING)
+      cmd_warp(NULL);
     update();
     save_history_point();
   }
@@ -1642,6 +1744,14 @@ void sigchld(int sig) {
   }
 }
 
+void sighup(int sig) {
+  restart();
+}
+
+void restart() {
+  execvp(g_argv[0], g_argv);
+}
+
 void recordings_save(const char *filename) {
   FILE *output = NULL;
   int i = 0;
@@ -1732,30 +1842,44 @@ void closepixel(Display *dpy, Window zone, mouseinfo_t *mouseinfo) {
 int main(int argc, char **argv) {
   char *pcDisplay;
   int ret;
+  const char *prog = argv[0];
+
+  g_argv = argv;
 
   if ((pcDisplay = getenv("DISPLAY")) == NULL) {
     fprintf(stderr, "Error: DISPLAY environment variable not set\n");
-    exit(1);
+    return EXIT_SUCCESS;
   }
 
   if ((dpy = XOpenDisplay(pcDisplay)) == NULL) {
     fprintf(stderr, "Error: Can't open display: %s\n", pcDisplay);
-    exit(1);
+    return EXIT_SUCCESS;
   }
 
   if (argc > 1 && (!strcmp(argv[1], "version") 
                    || !strcmp(argv[1], "-v") 
                    || !strcmp(argv[1], "--version"))) {
     printf("keynav %s\n", KEYNAV_VERSION);
-    return 0;
+    return EXIT_SUCCESS;
   }
 
   signal(SIGCHLD, sigchld);
+  signal(SIGHUP, sighup);
+  signal(SIGUSR1, sighup);
   xdo = xdo_new_with_opened_display(dpy, pcDisplay, False);
 
   parse_config();
   query_screens();
 
+  if (argc == 2) {
+    handle_commands(argv[1]);
+  } else if (argc > 2) {
+    fprintf(stderr, "Usage: %s [command string]\n", prog);
+    fprintf(stderr, "Did you quote your command string?\n");
+    fprintf(stderr, "  Example: %s 'loadconfig mykeynavrc,daemonize'\n", prog);
+    return EXIT_FAILURE;
+  }
+
   /* Sync with the X server.
    * This ensure we errors about XGrabKey and other failures
    * before we try to daemonize */
@@ -1764,14 +1888,7 @@ int main(int argc, char **argv) {
   if (daemonize) {
     printf("Daemonizing now...\n");
     daemon(0, 0);
-  }
-
-  if (argc == 2) {
-    handle_commands(argv[1]);
-  } else if (argc > 2) {
-    fprintf(stderr, "Usage: %s [command string]\n", argv[0]);
-    fprintf(stderr, "Did you quote your command string?\n");
-    exit(1);
+    is_daemon = True;
   }
 
   while (1) {
@@ -1783,11 +1900,12 @@ int main(int argc, char **argv) {
         handle_keypress((XKeyEvent *)&e);
         break;
 
+      /* MapNotify means the keynav window is now visible */
       case MapNotify:
         update();
         break;
 
-      // Map and Configure events mean the window was changed or is now mapped.
+      // Configure events mean the window was changed (size, property, etc)
       case ConfigureNotify:
         update();
         break;
@@ -1824,3 +1942,4 @@ int main(int argc, char **argv) {
 
   xdo_free(xdo);
 } /* int main */
+
